use crate::db::{db_get_latest_posts, db_new_post};
use crate::ServerState;

use actix_web::web::{Data, Query};
use actix_web::{get, post, web::Json, HttpResponse, Responder, Result};
use log::info;
use serde::{Deserialize, Serialize};
use sqlx::FromRow;

// The post as it is received from the client
// The token is used to identify the user
#[derive(Debug, Serialize, Deserialize)]
pub struct NewPost {
    pub content: String,
    pub token: String,
}

// The post as it is stored in the database, with all the related metadata
#[derive(Debug, Serialize, Deserialize, Clone, FromRow)]
pub struct Post {
    pub id: i64,
    pub user_id: i64,
    pub content: String,
    pub upvotes: i64,
    pub downvotes: i64,
    pub created_at: chrono::NaiveDateTime,
    pub updated_at: chrono::NaiveDateTime,
}

/// The user as it is stored in the database, with all the related metadata
#[derive(Debug, Serialize, Deserialize, Clone, FromRow)]
pub struct User {
    pub id: i64,
    pub username: String,
    pub password: String,
    pub created_at: chrono::NaiveDateTime,
    pub updated_at: chrono::NaiveDateTime,
}

// These look like /posts?limit=10&offset=20 in the URL
// Note that these are optional
/// Query parameters for the /posts endpoint
#[derive(Debug, Serialize, Deserialize)]
struct QueryParams {
    limit: Option<i64>,
    offset: Option<i64>,
}

/// Gets all posts from the database, query parameters are optional
/// If limit is not specified, it defaults to a sane value
#[get("/posts")]
pub async fn get_posts(
    query: Query<QueryParams>,
    state: Data<ServerState>,
) -> Result<impl Responder> {
    if let (Some(lim), Some(ofs)) = (query.limit, query.offset) {
        return Ok(HttpResponse::Ok()
            .json(db_get_latest_posts(&state.pool, std::cmp::min(lim, 30), ofs).await));
    }
    Ok(HttpResponse::Ok().json(db_get_latest_posts(&state.pool, 30, 0).await))
}

/// Creates a new post, requires a token in release mode
#[post("/posts")]
pub async fn new_post(new_post: Json<NewPost>, state: Data<ServerState>) -> Result<impl Responder> {
    return match db_new_post(new_post.into_inner(), &state.pool).await {
        Some(post) => {
            info!("Created post {:?}", post.id);
            Ok(HttpResponse::Ok().json(post))
        }
        None => Ok(HttpResponse::InternalServerError().json("Error")),
    };
}
